# -*- encoding: utf-8 -*-
##############################################################################
#
#    Copyright (C) 2011 Gábor Dukai <gdukai@gmail.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
import copy

from osv import fields, osv
from tools.translate import _

max_moves_on_picking = 35

class stock_wave(osv.osv):
    _name = 'stock.wave'

    _columns = {
        'name': fields.char('Reference', size=32, select=True),
        'state': fields.selection([
            ('in_progress', 'In Progress'),
            ('done', 'Done'),
        ], 'Status', readonly=True, select=True),
        'date': fields.datetime('Date Created'),
        'wave_picking_ids': fields.one2many('stock.wave.picking', 'wave_id',
            'Wave Pickings', readonly=True),
    }

    _defaults = {
        'name': lambda self, cr, uid, context: '/',
        'state': lambda *a: 'in_progress',
        'date': lambda *a: time.strftime('%Y-%m-%d %H:%M:%S'),
    }

    def create(self, cr, uid, vals, context=None):
        if ('name' not in vals) or (vals.get('name') == '/'):
            vals['name'] = self.pool.get('ir.sequence').get(cr, uid, 'stock.wave')
        return super(stock_wave, self).create(cr, uid, vals, context)

    def unlink(self, cr, uid, ids, context=None):
        for wave in self.browse(cr, uid, ids):
            for wpick in wave.wave_picking_ids:
                if wpick.state == 'done':
                    raise osv.except_osv(_('UserError'),
                        _("Can't delete an already done picking!"))
        return super(stock_wave, self).unlink(cr, uid, ids, context)

    def generate_picking_lines(self, cr, uid, picking_ids):
        move_obj = self.pool.get('stock.move')
        prod_obj = self.pool.get('product.product')
        move_ids = move_obj.search(cr, uid, [
            ('picking_id', 'in', picking_ids),
            ('state', 'not in', ['cancel', 'done']),
        ])
        moves = []
        for move in move_obj.browse(cr, uid, move_ids):
            moves.append({
                'name': move.product_id.name,
                'categ_id': move.product_id.categ_id.id,
                'id': move.id,
                'product_id': move.product_id.id,
                'uom_id': move.product_id.uom_id.id,
                'state': move.state,
            })
        prod_obj.sort_by_category(cr, uid, moves)
        assigned_lines = []
        other_lines = []
        line = {'product_id': None}
        any_move_assigned = False
        for move in moves:
            if line['product_id'] != move['product_id'] and line['product_id']:
                if any_move_assigned:
                    assigned_lines.append((0, 0, copy.copy(line)))
                else:
                    other_lines.append((0, 0, copy.copy(line)))
            if line['product_id'] != move['product_id']:
                line = {
                    'product_id': move['product_id'],
                    'uom_id': move['uom_id'],
                    'move_ids': [(6, 0, [])]
                }
                any_move_assigned = False
            if line['product_id'] == move['product_id']:
                line['move_ids'][0][2].append(move['id'])
                if move['state'] == 'assigned':
                    any_move_assigned = True
        if any_move_assigned:
            assigned_lines.append((0, 0, line))
        else:
            other_lines.append((0, 0, line))
        return (assigned_lines, other_lines)

    def create_from_picking(self, cr, uid, picking_ids):
        wpick_obj = self.pool.get('stock.wave.picking')
        pick_obj = self.pool.get('stock.picking')
        for pick in pick_obj.browse(cr, uid, picking_ids):
            if pick.wave_id:
                raise osv.except_osv(_('UserError'),
                    _('The picking %s is already in a wave!') % pick.name)
            if pick.type not in ('out', 'delivery', 'internal'):
                raise osv.except_osv(_('UserError'),
                    _('Only outgoing and internal shipments can be picked in a wave!'))
        wave_id = self.create(cr, uid, {})
        (assigned_lines, other_lines) \
            = self.generate_picking_lines(cr, uid, picking_ids)
        count_moves = 0
        count_pages = 1
        batched_lines = []
        wpick_ids = []
        for line in assigned_lines + other_lines:
            if count_moves + len(line[2]['move_ids'][0][2]) \
                > max_moves_on_picking:
                wpick_ids.append(wpick_obj.create(cr, uid, {
                    'wave_id': wave_id,
                    'name': str(count_pages),
                    'line_ids': batched_lines,
                }))
                batched_lines = []
                count_moves = 0
                count_pages += 1
            batched_lines.append(line)
            count_moves += len(line[2]['move_ids'][0][2])
        if batched_lines:
            wpick_ids.append(wpick_obj.create(cr, uid, {
                'wave_id': wave_id,
                'name': str(count_pages),
                'line_ids': batched_lines,
            }))
        pick_obj.write(cr, uid, picking_ids, {'wave_id': wave_id})
        return wpick_ids

stock_wave()

class stock_wave_picking(osv.osv):
    _name = 'stock.wave.picking'

    _columns = {
        'wave_id': fields.many2one('stock.wave', 'Wave', required=True,
            ondelete='cascade'),
        'name': fields.char('Reference', size=32, select=True),
        'state': fields.selection([
            ('in_progress', 'In Progress'),
            ('done', 'Done'),
        ], 'Status', readonly=True, select=True),
        'line_ids': fields.one2many('stock.wave.picking.line', 'wave_picking_id',
            'Lines', readonly=True),
    }

    _defaults = {
        'name': lambda self, cr, uid, context: '/',
        'state': lambda *a: 'in_progress',
    }

    def unlink(self, cr, uid, ids, context=None):
        for wpick in self.browse(cr, uid, ids):
            if wpick.state == 'done':
                raise osv.except_osv(_('UserError'),
                    _("Can't delete an already done picking!"))
        return super(stock_wave_picking, self).unlink(cr, uid, ids, context)

stock_wave_picking()

class stock_wave_picking_line(osv.osv):
    _name = 'stock.wave.picking.line'

    def get_quantity(self, cr, uid, ids, field_names=None, arg=False,
        context=None):
        uom_obj = self.pool.get('product.uom')
        res = {}
        for line in self.browse(cr, uid, ids):
            res[line.id] = 0
            for move in line.move_ids:
                res[line.id] += uom_obj._compute_qty_obj(cr, uid,
                    move.product_uom, move.product_qty, line.uom_id)
        return res

    _columns = {
        'wave_picking_id': fields.many2one('stock.wave.picking',
            'Wave Picking', required=True, ondelete='cascade'),
        'product_id': fields.many2one('product.product', 'Product',
            required=True),
        'uom_id': fields.many2one('product.uom', 'UoM', required=True),
        'quantity': fields.function(get_quantity, method=True,
            type='float', string='Quantity'),
        'move_ids': fields.many2many('stock.move', 'wave_picking_move_rel', \
            'wave_picking_ids', 'move_ids', 'Stock Moves', required=True),
    }

stock_wave_picking_line()